/*
 * Asterisk -- A telephony toolkit for Linux.
 *
 * Determine route
 * 
 * Copyright (C) 1999-2004, Digium, Inc.
 *
 * Mark Spencer <markster@digium.com>
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License
 */

#include <asterisk/lock.h>
#include <asterisk/file.h>
#include <asterisk/logger.h>
#include <asterisk/channel.h>
#include <asterisk/pbx.h>

#include <asterisk/utils.h>
#include <asterisk/module.h>
#include <asterisk/app.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>

#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>


static int rr;

typedef struct {

	char host[128];
	int active;

} iax_routes;


static int get_route(char *route);
static int get_routes(iax_routes *routes, int max);
static int udpping(char *host, int port, int timeout_usec);



AST_MUTEX_DEFINE_STATIC(rrlock);

static char *tdesc = "Determine route to next call processor route and store the dialstring in env variable RR_DIALSTRING";

static char *app = "RoundRobinRoute";

static char *synopsis = "Determine route to call processor in a round robin manner";

static char *descrip =
"  RoundRobinRoute(): Determine route to next available call processor in \n"
"a round robin manner. If  the route detection fails, and there exists a \n"
"priority n + 101, where 'n' is the priority of the current instance, then  \n"
"the  channel  will  be  setup  to continue at that priority level.  Otherwise,\n"
"RoundRobinRoute returns 0.\n";

STANDARD_LOCAL_USER;

LOCAL_USER_DECL;


static int udpping(char *host, int port, int timeout_usec) {
   int rc = 0;
   int r, s;
   char  msg[1024];

   struct sockaddr_in destaddr;

   fd_set rfds;
   struct timeval timeout;

   s = socket(PF_INET, SOCK_DGRAM, 17);

   memset(&destaddr, 0, sizeof(destaddr));

   destaddr.sin_port = htons(port);
   destaddr.sin_addr.s_addr = inet_addr(host);

   strcpy(msg, "\x80\x00\x00\x00\x00\x00\x00\x00\x00\x00\x06\x1e");

   timeout.tv_sec = 0;
   timeout.tv_usec = timeout_usec;

   FD_ZERO(&rfds);
   FD_SET(s, &rfds);

   sendto(s, msg, 12, MSG_NOSIGNAL, (struct sockaddr *) &destaddr, sizeof(destaddr) );

   r = select(s + 1, &rfds, NULL, NULL, &timeout);

   if (r) {
      r = recv(s, msg, 1024, 0);
      if (r) rc = 1;
   }

   return rc;
}

static int get_routes(iax_routes *routes, int max) { 

	FILE *f;
	int count = 0;
	char buf[128];
	int l;

	f = fopen("/etc/asterisk/routes", "r");
	if (!f) return -1;

	while (fgets(buf, sizeof(buf) - 1, f)) {
		l = strlen(buf);

		if (l && buf[l - 1] == '\n') buf[--l] = 0;
		if (l == 0 || buf[0] == '#' ) continue;


		routes[count].active = 1;
		strcpy(routes[count].host, buf);
		count += 1;

		if (count == max - 1) break;

	}
	fclose(f);

	return count;
}

static int get_route(char *route) {

	iax_routes routes[32];

	int rc = -1;
	int i, count;

	count = get_routes(routes, 32);

	if (count < 1) {
		ast_verbose("RoundRobinRoute: No routes available!!");
		return -1;
	}

   ast_mutex_lock(&rrlock);
	while (1) {

		if (++rr == count) rr = 0;

		if (udpping(routes[rr].host, 4569, 250000)) {

			if (!routes[rr].active) {
				routes[rr].active = 1;
				ast_verbose("RoundRobinRoute: %s route reactivated\n", routes[rr].host);
			}
			rc = 1;

			strcpy(route, routes[rr].host);
			break;

		} else {
			routes[rr].active = 0;
		}

		if (!routes[rr].active) {
			ast_verbose("RoundRobinRoute: %s route failed\n", routes[rr].host);
		}

	}
   ast_mutex_unlock(&rrlock);

	return rc;
}


static int roundrobin_exec_helper(struct ast_channel *chan, void *data, int failmode)
{
	int res=0;
	struct localuser *u;
	char route[128];

	LOCAL_USER_ADD(u);

	if (get_route(route) == -1) {
		ast_log(LOG_WARNING, "Unable to get route!");

		if ( ast_exists_extension(chan, chan->context, chan->exten, chan->priority + 101, chan->callerid)) {
			chan->priority += 100;
			return 0;
		}
	}
	pbx_builtin_setvar_helper(chan, "RR_DIALSTRING", route);
	return res;
}

static int roundrobin_exec(struct ast_channel *chan, void *data)
{
	return roundrobin_exec_helper(chan, data, 0);
}



int unload_module(void)
{
	STANDARD_HANGUP_LOCALUSERS;
	return ast_unregister_application(app);
}

int load_module(void)
{
	rr = 0;

	return ast_register_application(app, roundrobin_exec, synopsis, descrip);
}

char *description(void)
{
	return tdesc;
}

int usecount(void)
{
	int res;
	STANDARD_USECOUNT(res);
	return res;
}

char *key()
{
	return ASTERISK_GPL_KEY;
}
